            // Slave for a simple UDP file transfer server.
            // The protocol is modelled on the Trivial File Transfer
            // Protocol (TFTP) described in RFC 1350.
            // Only sending of a file is implemented, and the
            // transfer mode is binary (i.e. octet).

import java.net.*;
import java.io.*;

        // Handle simple file transfer to a single client.
class UDPFileTransferSlave extends UDPUser implements Runnable {
        // The requested filename is encoded in the packet.
    public UDPFileTransferSlave(DatagramPacket request) throws SocketException {
        super(request.getAddress());
        // Keep track of the client's port.
        clientPort = request.getPort();
        // Extract the name of the file to be transferred.
        setFilename(request.getData(),request.getLength());
    }

        // Implementation of Runnable.
    public void run(){
        // Create space for a DATA packet.
        byte[] data = new byte[UDPFileTransfer.fullDataPacket];
        DatagramPacket packet = new DatagramPacket(data,data.length,
                                       getCorrespondent(),getClientPort());

        // Check that the filename is acceptable.
        String filename = getFilename();
        if(!filenameOk(filename)){
            // Reject it.
            System.out.println("Filename rejected: "+filename);
            sendError(UDPFileTransfer.AccessViolation,packet);
        }
        else{
            // Open the file.
            InputStream inStream = null;
            try{
                inStream = openFile(filename,packet);
                transfer(inStream,packet);
            }
            catch(Exception e){
                System.err.println("Unable to transfer: "+filename+
                                   " for reason: "+e.getMessage());
            }
            finally{
                if(inStream != null){
                    try{
                        inStream.close();
                    }
                    catch(IOException e){
                    }
                }
            }
        }
        close();
    }

        // Open the given file and return an InputStream reference to it.
        // Send an ERROR packet if the file cannot be opened for
        // some reason.
    protected InputStream openFile(String filename,DatagramPacket packet)
                                throws IOException {
        try{
            return new FileInputStream(filename);
        }
        catch(IOException e){
            // Failed for some reason. Let the client know.
            System.out.println("Filename could not be opened: "+filename);
            sendError(UDPFileTransfer.FileNotFound,packet);
            throw e;
        }
    }

        // Perform the task of transferring the open file.
        // Receive an ACK packet for each DATA packet sent.
        // Be prepared to resend a DATA packet if an appropriate
        // ACK is not received.
    protected void transfer(InputStream inStream,DatagramPacket packet)
                                throws Exception {
        System.out.println("Transferring: "+getFilename());
        // Create a packet to receive ACKs.
        byte[] ack = new byte[UDPFileTransfer.opcodeSize+
                              UDPFileTransfer.blockNumberSize];
        DatagramPacket ackPacket = new DatagramPacket(ack,ack.length);

        // Keep track of the next block to send.
        int nextBlockNumber = 1;
        // Prepare the first block of data to send.
        byte[] data = packet.getData();
        twoByteEncode(UDPFileTransfer.DATA,data,0);
        // Read the first block of data.
        int bytesRead = inStream.read(data,UDPFileTransfer.dataHeaderLength,
                                      UDPFileTransfer.MaxDataPacket);
        boolean endOfTransfer = false;
        while(!endOfTransfer){
            // Encode the block number.
            twoByteEncode(nextBlockNumber,data,UDPFileTransfer.opcodeSize);
            // Set the length of the packet.
            packet.setLength(UDPFileTransfer.dataHeaderLength+bytesRead);
            // Send the packet and expect an ACK.
            sendAndReceive(packet,ackPacket);

            // Check the ACK.
            int length = ackPacket.getLength();
            if(!lengthOk(length)){
                throw new Exception("Bad length packet received.");
            }
            else if(!opcodeOk(ack)){
                throw new Exception("Non-ACK packet received.");
            }
            else if(expectedBlock(ack,nextBlockNumber)){
                System.out.println("Received ACK for block: "+nextBlockNumber);
                // Check what was sent last time.
                if(bytesRead == UDPFileTransfer.MaxDataPacket){
                    // There should be more to send.
                    bytesRead = inStream.read(data,
                                        UDPFileTransfer.dataHeaderLength,
                                        UDPFileTransfer.MaxDataPacket);
                    nextBlockNumber++;
                }
                else{
                    System.out.println("Transfer complete.");
                    endOfTransfer = true;
                }
            }
            else{
                // Resend the last acknowledgement.
                // Warning: this results in the "Sorcerer's Apprentice
                // Syndrome" (see RFC 1123).
                System.out.println("Resend.");
            }
        }
    }

        // Send an error indication to the client.
    protected void sendError(int errorCode,DatagramPacket packet){
        try{
            byte[] data = packet.getData();
            twoByteEncode(UDPFileTransfer.ERROR,data,0);
            twoByteEncode(errorCode,data,UDPFileTransfer.opcodeSize);
            DatagramSocket socket = getSocket();
            socket.send(packet);
        }
        catch(IOException e){
            System.err.println("Error send failed.");
        }
    }

        // Is the filename ok?
        // Reject anything with '/' or '\\' characters.
    protected boolean filenameOk(String filename){
        return (filename != null) &&
               (filename.indexOf('/') < 0) &&
               (filename.indexOf('\\') < 0);
    }

        // Make sure the received data contains at least an opcode
        // and block number.
    protected boolean lengthOk(int length){
        return length >= (UDPFileTransfer.opcodeSize+
                          UDPFileTransfer.blockNumberSize);
    }

        // Make sure the received data is tagged as DATA.
    protected boolean opcodeOk(byte[] data){
        return (data[0] == 0) && (data[1] == UDPFileTransfer.ACK);
    }

        // Make sure the received data is the right block number.
    protected boolean expectedBlock(byte[] data, int expected){
        int offset = UDPFileTransfer.opcodeSize;
        byte highByte = (byte) ((expected >> 8) & 0x0f);
        byte lowByte = (byte) (expected & 0x0f);
        return (data[offset] == highByte) && (data[offset+1] == lowByte);
    }

        // Store the lower order two bytes of value into data.
    protected void twoByteEncode(int value,byte[] data,int where){
        data[where] = (byte) ((value >> 8) & 0x0f);
        data[where+1] = (byte) (value & 0x0f);
    }

    protected String getFilename(){
        return filename;
    }

    protected void setFilename(byte[] data,int length){
        // The length of the filename omits the opcode and 0 byte
        // in data.
        int filenameLength = length-(UDPFileTransfer.opcodeSize+1);
        if(filenameLength > 0){
            byte[] name = new byte[filenameLength];
            System.arraycopy(data,UDPFileTransfer.opcodeSize,
                                name,0,filenameLength);
            filename = new String(name);
        }
        else{
            filename = null;
        }
    }

    protected int getClientPort(){
        return clientPort;
    }

    private final int clientPort;
    // The file to be transferred.
    private String filename;
}
